{-# LANGUAGE OverloadedStrings #-}

module Main where

import Data.Aeson
import Data.String
import Data.Maybe (maybe)
import Control.Concurrent.MVar
import Control.Monad (forM_)
import qualified Network.WebSockets as WS
import qualified Network.Wai.Handler.WebSockets as WS
import qualified Network.Wai.Application.Static as Static
import qualified WaiAppStatic.Types as Static
import qualified Network.Wai.Handler.Warp as Warp
import System.Random (newStdGen)
import Messages
import OT
import Control.Exception (catch, throw)

data ServerState = ServerState {
  ot :: Server TextOperation [Char],
  idGen :: ClientId,
  clients :: [(ClientId,WS.Connection)]
}

type ClientId = Int

initialState = ServerState {
  ot = (Server ("" :: [Char]) ([] :: [TextOperation])),
  idGen = 0,
  clients = []
}

handleSocket :: WS.Connection -> ClientId -> Int -> MVar ServerState -> IO ()
handleSocket conn id offset server = do
  -- Empfange die Daten vom Client
  let disconnect err = do
        putStrLn $ "disconnected:  client " ++ show id
        modifyMVar_ server $ \s ->
          return s { clients = filter ((/= id). fst) (clients s)  }
        throw (err :: WS.ConnectionException)
  msg <- catch (WS.receiveData conn) disconnect
  case decode msg of
    Just (ClientEdit rev op) -> do
      op' <- modifyMVar server $ \s -> do
        let (op',ot') = appendOperation (ot s) op (offset + rev)
        forM_ (clients s) $ \(cid,client) ->
          if cid /= id
          then WS.sendTextData client (encode (RemoteEdit op'))
          else WS.sendTextData client (encode Ack)
        return (s { ot = ot'},op')
      handleSocket conn id offset server
    Nothing -> do
      print msg
      error "could not decode client message"

-- Der statische Teil der Web Anwendung (html + javascript)
static = Static.staticApp (staticSettings)
staticSettings = (Static.defaultFileServerSettings "client/assets")
  -- Caching ausschalten, damit die Anwendung nicht neu gestartet werden
  -- muss, wenn der Scala teil neu Kompiliert wurde.
  { Static.ssMaxAge = Static.NoMaxAge }

-- Der WebSocket
socket server pending = do
  conn <- WS.acceptRequest pending
  WS.forkPingThread conn 30
  (id,offset) <- modifyMVar server $ \s -> do
    let id = 1 + idGen s
    let clients' = (id,conn) : clients s
    let offset = 1 + serverRevision (ot s)
    let stateOp = serverStateOp (ot s)
    WS.sendTextData conn (encode (RemoteEdit stateOp))
    return (s {idGen = id, clients = clients'}, (id,offset))
  putStrLn $ "connected: client " ++ show id
  handleSocket conn id offset server

-- Starte die Webanwendung auf Port 3000
main = do
  server <- newMVar $ Main.initialState
  Warp.run 3000 $
    WS.websocketsOr WS.defaultConnectionOptions (socket server) static
